-- $Id: DualSpeccer.lua 7 2009-04-18 00:48:49Z jeffrey0 $
DualSpeccer = LibStub("AceAddon-3.0"):NewAddon("DualSpeccer", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

local filterTimer = nil

local groupNames = {"primary", "secondary"}
local groupNamesCaps = {"Primary", "Secondary"}

local primarySpecSpellName = GetSpellInfo(63645)
local secondarySpecSpellName = GetSpellInfo(63644)

local specCache = {}

local rightClickMenu = {
	{
		text = "Hide switch spam",
		func = function() DualSpeccer.db.profile.filterSpam = not DualSpeccer.db.profile.filterSpam end,
		checked = function() return DualSpeccer.db.profile.filterSpam end,
		tooltipTitle = "Hide switch spam",
		tooltipText = "Hides the spell learned/unlearned messages when swapping specs."
	}
}

-- Match patterns we use. Replace any "." with "%." and any "%s" with "(.*)".
local spamFilterMatch1 = string.gsub(ERR_LEARN_ABILITY_S:gsub('%.', '%.'), '%%s', '(.*)')
local spamFilterMatch2 = string.gsub(ERR_LEARN_SPELL_S:gsub('%.', '%.'), '%%s', '(.*)')
local spamFilterMatch3 = string.gsub(ERR_SPELL_UNLEARNED_S:gsub('%.', '%.'), '%%s', '(.*)')

-- Localize some common functions.
local GetNumTalentGroups, GetActiveTalentGroup = GetNumTalentGroups, GetActiveTalentGroup

--[[
	TODO:
	- Some settings on controlling the text that appears in the Broker launcher.
	- Localization (AceLocale?)
]]

--[[
	Information on zoning and talent info:
	
	On initial login, proper talent information isn't available until an onUpdate (as far as I've seen).
	
	On zoning to a different server, the following happens:
	- PLAYER_LEAVING_WORLD
	- Talent data becomes unavailable (returns 0/0/0)
	- PLAYER_TALENT_UPDATE
	- PLAYER_ENTERING_WORLD and related spam (BAG_UPDATE events etc)
	- Talent data becomes available
	- UPDATE_INSTANCE_INFO
	As such, we unregister PLAYER_TALENT_UPDATE between LEAVING and ENTERING.
	
	Teleporting around on the same continent doesn't interest us and nothing we care about fires.
]]

--[[
	What happens when switching specs:
	- UNIT_SPELLCAST_SENT player Activate Primary/Secondary Spec
	- UNIT_SPELLCAST_START player Activate Primary/Secondary Spec
	- CHAT_MSG_SYSTEM and SPELLS_CHANGED spam "You have unlearned blah"
	- CHAT_MSG_SYSTEM and LEARNED_SPELL_IN_TAB and SPELLS_CHANGED spam "You have learned blah"
	- UNIT_SPELLCAST_FAILED_QUIET player Activate Primary/Secondary Spec
	- UNIT_SPELLCAST_STOP player Activate Primary/Secondary Spec
	- PLAYER_TALENT_UPDATE
	- ACTIVE_TALENT_GROUP_CHANGED newGroupNumber, oldGroupNumber
	- ACTIONBAR_SLOT_CHANGED spam
	- UNIT_SPELLCAST_SUCCEEDED player Activate Primary/Secondary Spec
	- GLYPH_UPDATED/ADDED, SPELL_USABLE info and what not.
	
	Yes, we fail and then succeed at casting. I didn't make it...
]]

DualSpeccerLDB = LibStub("LibDataBroker-1.1"):NewDataObject("DualSpeccer", {
	type = "data source",
	text = "DualSpeccer",
	icon = "Interface\\Icons\\Ability_Marksmanship",
	tocname = "DualSpeccer"
})

function DualSpeccer:OnInitialize()
	local dbDefaults = {
		profile = {
			filterSpam = true
		}
	}
	self.db = LibStub("AceDB-3.0"):New("DualSpeccerDB", dbDefaults, "Default")
end

function DualSpeccer:OnEnable()
	self:RegisterEvent("UNIT_SPELLCAST_START")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	self:RegisterEvent("PLAYER_LEAVING_WORLD")
	
	-- Chances are good that the P_E_W will fire after we're done, registering PLAYER_TALENT_UPDATE twice. This
	-- is more of an AceAddon implementation thing and CallbackHandler overwrites it anyway.
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TALENT_UPDATE")
	
	-- The important part here is that it happens after the first OnUpdate which is seemingly the time when correct
	-- information becomes available. There's a kicker though: on a ReloadUI, our information is proper right away.
	self:ScheduleTimer("Update", 0.01)
end

function DualSpeccer:OnDisable()
	
end

-- A "corrupted" talent update fires between LEAVING and ENTERING when zoning. We don't want this one.
function DualSpeccer:PLAYER_LEAVING_WORLD(event)
	self:UnregisterEvent("PLAYER_TALENT_UPDATE")
end

function DualSpeccer:PLAYER_ENTERING_WORLD(event)
	self:RegisterEvent("PLAYER_TALENT_UPDATE")
end

function DualSpeccer:ACTIVE_TALENT_GROUP_CHANGED(event, ...)
	local newGroup = ...
	if filterTimer then
		self:StopFilter()
	end
	self:Print(("Switched to %s talent spec."):format(groupNames[newGroup]))
end

function DualSpeccer:UNIT_SPELLCAST_START(event, ...)
	local unit, spellName = ...
	if unit == "player" and (spellName == primarySpecSpellName or spellName == secondarySpecSpellName) then
		-- We are now switching specs. Hide learned messages for the next ten seconds.
		self:StartFilter()
	end
end

function DualSpeccer:UNIT_SPELLCAST_INTERRUPTED(event, ...)
	local unit, spellName = ...
	if unit == "player" and (spellName == primarySpecSpellName or spellName == secondarySpecSpellName) then
		-- We moved (or something). End filtering.
		self:StopFilter()
	end
end

function DualSpeccer:PLAYER_TALENT_UPDATE(event)
	self:Update()
end

-- Updates our talent information.
function DualSpeccer:Update()
	for i = 1, GetNumTalentGroups() do
		specCache[i] = specCache[i] or {}
		local thisCache = specCache[i]
		TalentFrame_UpdateSpecInfoCache(thisCache, false, false, i)
		if thisCache.primaryTabIndex and thisCache.primaryTabIndex ~= 0 then
			thisCache.specName = thisCache[thisCache.primaryTabIndex].name
			thisCache.mainTabIcon = thisCache[thisCache.primaryTabIndex].icon
		else
			thisCache.specName = "Hybrid"
			thisCache.mainTabIcon = "Interface\\Icons\\Ability_Marksmanship"
		end
		thisCache.specGroupName = groupNamesCaps[i]
	end
	
	local activeGroupNum = GetActiveTalentGroup()
	local curCache = specCache[activeGroupNum]
	
	DualSpeccerLDB.text = curCache.specGroupName
	DualSpeccerLDB.icon = curCache.mainTabIcon
end

-- Our chat filter function.
local function filter(self, event, msg)
	-- You have unlearned Blahblah.
	-- You have learned a new spell: Blahblah.
	-- You have learned a new ability: Blahblah.
	if (msg:find(spamFilterMatch1) ~= nil or msg:find(spamFilterMatch2) ~= nil or msg:find(spamFilterMatch3) ~= nil) then
		return true
	end
end

-- Stops filtering chat.
function DualSpeccer:StopFilter()
	if not filterTimer then return end
	self:CancelTimer(filterTimer, true)
	filterTimer = nil
	ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SYSTEM", filter)
end

-- Starts filtering chat.
function DualSpeccer:StartFilter()
	if not self.db.profile.filterSpam or filterTimer then return end
	filterTimer = self:ScheduleTimer("StopFilter", 10)
	ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", filter)
end

-- Menu anchoring stuff stolen from ClosetGnome.
local menuPopupFrame = CreateFrame("Frame", "DualSpeccerMenu", UIParent, "UIDropDownMenuTemplate")
function DualSpeccerLDB.OnClick(displayFrame, button)
	button = button or "LeftButton"
	
	if button == "LeftButton" then
		if GetNumTalentGroups() < 2 then
			DualSpeccer:Print("This character does not have a second specialization.")
			return
		end
		
		local curSpec = GetActiveTalentGroup()
		-- -1 + 3 = 2 and -2 + 3 = 1.
		local newSpec = -curSpec + 3
		SetActiveTalentGroup(newSpec)
	elseif button == "RightButton" then
		EasyMenu(rightClickMenu, menuPopupFrame, displayFrame, 20, 4, "MENU")
	end
end

function DualSpeccerLDB.OnTooltipShow(tooltip)
	tooltip:AddLine("DualSpeccer")
	local curGroup = GetActiveTalentGroup()
	local curCache = specCache[curGroup]
	
	if not curCache then
		DualSpeccer:Update()
		curCache = specCache[curGroup]
		if not curCache then
			tooltip:AddLine("Something really bad happened.")
			return
		end
	end
	
	tooltip:AddDoubleLine("Active spec", ("%s (%s)"):format(curCache.specGroupName, curCache.specName))
	
	-- We're showing the tooltip while our data is bad. Just don't show the spec info.
	if not specCache[1][1] or specCache[1][1].pointsSpent == nil then
		return
	end
	
	-- Greenish if selected, yellowish white if not selected.
	local primSpecColor = "ffff99"
	if curGroup == 1 then primSpecColor = "6adb54" end
	
	tooltip:AddDoubleLine("Primary spec", ("|cff%s%s|r (%d/%d/%d)"):format(primSpecColor, specCache[1].specName, specCache[1][1].pointsSpent, specCache[1][2].pointsSpent, specCache[1][3].pointsSpent))
	if GetNumTalentGroups() >= 2 then
		local secSpecColor = "ffff99"
		if curGroup == 2 then secSpecColor = "6adb54" end
		tooltip:AddDoubleLine("Secondary spec", ("|cff%s%s|r (%d/%d/%d)"):format(secSpecColor, specCache[2].specName, specCache[2][1].pointsSpent, specCache[2][2].pointsSpent, specCache[2][3].pointsSpent))
		tooltip:AddLine(" ")
		tooltip:AddLine("|cffeda55fClick|r to switch to your other specialization.")
		tooltip:AddLine("|cffeda55fRight-Click|r to open the menu.")
	end
end

